/*
# Planète Casio v5 shoutbox

This script contains all of the client-side logic for the Planète Casio
shoutbox. Unlike the legacy v4 shoutbox which was a poll-based HTTP service,
this shoutbox is run by an IRC server and this is mostly an IRC client.

Support for IRC features is fairly limited as the shoutbox has never been a
very fancy chat. For full functionality, use a standard IRC client.

The one unusual feature in this program is the interaction with the Planète
Casio website. This solves an issue of permissions that is difficult to manage
with pure IRC. The expectation for the shoutbox is that:
- Anyone can read but only community members can write;
- It should load immediately when embedded into website pages and allow logged
  in users to read and write messages with no further interactions.

In the v4's cursed design, the shoutbox was part of the website API, so users
logged into their Planète Casio account would be recognized by the shoutbox
server (which is just the website) and authenticated by their session cookie.
This doesn't work with the IRC server, since the connections are very different
and cookies are not a reliable method of authentication. With a standard IRC
client, users would need to type in their password to connect to the chat
despite already being logged in to the website.

The chosen solution is as follows:
- Every embedded shoutbox connects to the IRC server as a read-only guest;
- Sending a message is done via a website request. The website identifies
  logged in users by their session cookies and forwards the messages to the
  IRC server locally;
- Messages from the website to the IRC server are attributed to their original
  authors by whichever nick/plugin trick works best.

Hence, this client supports different connection modes depending on the
authentication setup: normal IRC (with password), PCv4, PCv5.
(WIP at time of writing)
*/

"use strict";

/* WebSocket IRC with HTTP-based website interface. */
let irc = new function() {

  const State = Object.freeze({
    /* Websocket is disconnected. */
    DISCONNECTED: Symbol("DISCONNECTED"),
    /* We just opened a socket to the IRC server and have yet to register. */
    CONNECTING: Symbol("CONNECTING"),
    /* We are authenticated and ready to use the server. */
    READY: Symbol("READY"),
  });

  let conn = {
    /* IRC server's websocket URL (wss://...). */
    serverUrl:  undefined,
    /* The websocket connected to serverUrl */
    socket:     undefined,
    /* HTTP-based message post URL. Undefined when using pure IRC, or a v4/v5
       URL when using website authentication. */
    postUrl:    undefined,
    /* Current connection state. */
    state:      State.DISCONNECTED,
    /* Username with which we are logged in. Undefined when DISCONNECTED. */
    username:   undefined,
    /* Current channel. This remains defined even when DISCONNECTED. */
    channel:    undefined,
    /* Password for connection. Erased after authenticating. */
    password:   undefined,
    /* Capabilities enabled. */
    caps:       [],
  };

  let batches = new Map();

  /* List of channels and their information */
  this.channels = new Map();

  /*** Accessors/mutators ***/

  this.isConnected = function() {
    return conn.state !== State.DISCONNECTED;
  }
  this.isRunning = function() {
    return conn.state === State.READY;
  }
  this.stateString = function() {
    switch(conn.state) {
    case State.DISCONNECTED:    return "Disconnected";
    case State.CONNECTED:       return "Connecting...";
    case State.READY:           return "Connected";
    }
  }

  this.currentChannel = function() {
    return conn.channel;
  }
  this.selectChannel = function(name) {
    conn.channel = name;
  }
  this.isOnRemoteChannel = function() {
    const c = conn.channel;
    return (c !== undefined) && (c.startsWith("&") || c.startsWith("#"));
  }

  /*** Overridable hooks and callbacks ***/

  this.onLog = function(message) {};
  this.onStateChanged = function() {};
  this.onAuthenticated = function() {};
  this.onNewMessage = function(channel, date, author, message) {};
  this.onChannelChanged = function(channel) {};

  function log(...args) {
    irc.onLog(...args);
  }

  /*** Connection management ***/

  function setState(state) {
    conn.state = state;
    irc.onStateChanged();
  }

  this.connect = function(serverUrl, postUrl, username, password) {
    conn.serverUrl = serverUrl;
    conn.postUrl = postUrl;
    conn.socket = new WebSocket(conn.serverUrl);
    log("[v5shoutbox] Connecting...");

    conn.socket.onopen = function() {
      log("[v5shoutbox] Connected.");
      setState(State.CONNECTED);
      conn.username = username;
      conn.password = password;
      startExchange();
    }
    conn.socket.onmessage = function(net) {
      log("[<] " + net.data);
      let msg = new Message(net.data);
      if(msg.command === undefined) {
        log("[v5shoutbox] invalid message");
        return;
      }
      processMessage(msg);
    }
    conn.socket.onclose = function() {
      log("[v5shoutbox] Disconnected.");
      setState(State.DISCONNECTED);
    }
  }

  function sendRaw(command) {
    log("[>] " + command);
    conn.socket.send(command);
  }

  function send(command, args) {
    if(args.slice(0,-1).some(x => x.includes(" ")))
      return log("invalid spaces in non-last args: " + command.toString() +
                 " " + args.toString());

    if(args.length && args[args.length - 1].includes(" "))
      args[args.length - 1] = ":" + args[args.length - 1];

    sendRaw(command + " " + args.join(" "));
  }

  /*** IRC protocol ***/

  /* Regex for parsing messages */
  const rMessage = (function(){
    const rTags     = /(?:@([^ ]+) +)?/;
    const rSource   = /(?::([^ ]+) +)?/;
    const rCommand  = /(?:([a-zA-Z]+)|(\d{3}))/;
    const rArgs     = /((?: +[^\r\n :][^\r\n ]*)*)/;
    const rTrailing = /(?: +:([^\r\n]*))?/;
    return new RegExp([/^/, rTags, rSource, rCommand, rArgs, rTrailing, /$/]
                      .map(r => r.source).join(""));
  })();

  /* https://stackoverflow.com/questions/30106476 */
  function base64_encode(str) {
    return btoa(encodeURIComponent(str).replace(/%([0-9A-F]{2})/g,
        function toSolidBytes(match, p1) {
            return String.fromCharCode('0x' + p1);
    }));
  }

  class Message {
    constructor(text) {
      const matches = rMessage.exec(text);
      if(matches === undefined)
        return;

      this.tags = new Map();
      (matches[1] || "").split(";").filter(s => s).forEach(s => {
        const parts = s.split(/=(.*)/);
        const value = (parts[1] || "").replaceAll("\\\\", "\\")
                                      .replaceAll("\\:", ";")
                                      .replaceAll("\\r", "\r")
                                      .replaceAll("\\n", "\n")
                                      .replaceAll(/\\./g, m => m[1]);
        this.tags.set(parts[0], value);
      });

      this.source = matches[2];
      this.command = matches[3] || parseInt(matches[4]);
      this.args = (matches[5] || "").split(" ").filter(s => s);
      if(matches[6] !== undefined)
        this.args.push(matches[6]);
    }
  }
  this.Message = Message;

  function startExchange() {
    sendRaw("CAP LS 302");
    send("NICK", [conn.username]);
    send("USER", [conn.username, "*", "0", conn.username]);
  }

  function processMessage(msg) {
    if(msg.command === "PING") {
      send("PONG", msg.args);
      return;
    }
    if(msg.command === "BATCH") {
      const [mode, id] = [msg.args[0][0], msg.args[0].substring(1)];
      if(mode == "+")
        batches.set(id, []);
      batches.get(id).push(msg);
      if(mode == "-") {
        processBatch(batches.get(id));
        batches.delete(id);
      }
      return;
    }
    if(msg.tags.has("batch")) {
      batches.get(msg.tags.get("batch")).push(msg);
      return;
    }

    if(msg.command === "CAP" && msg.args[1] === "LS") {
      const caps = msg.args[2].split(" ").filter(x => x);
      const wishlist =
        ["message-tags", "server-time", "echo-message",
         "batch", "sasl", "draft/chathistory"];

      let req = wishlist.filter(x => caps.includes(x));
      if(caps.some(x => x.startsWith("sasl=")))
        req.push("sasl");

      if(req.length)
        send("CAP", ["REQ", req.join(" ")]);
      else {
        conn.caps = [];
        sendRaw("CAP END");
      }
    }

    if(msg.command === "CAP" && msg.args[1] === "ACK") {
      conn.caps = msg.args[2].split(" ").filter(x => x);
      if(conn.caps.includes("sasl"))
        sendRaw("AUTHENTICATE PLAIN");
      else
        sendRaw("CAP END");
    }
    if(msg.command === "AUTHENTICATE" && msg.args[0] === "+") {
      const arg = conn.username + "\0" + conn.username + "\0" + conn.password;
      const b64 = base64_encode(arg);
      log("[v5shoutbox] AUTHENTICATE command sent (not shown)");
      conn.socket.send("AUTHENTICATE " + b64);
    }

    /* Default authentication if SASL is unavailable or fails */
    if(msg.command === "NOTICE" && msg.args[1].includes("/AUTH")) {
      log("[v5shoutbox] AUTH command sent (not shown)");
      conn.socket.send("AUTH " + conn.username + ":" + conn.password);
    }

    if(msg.command === 900) {
      log("[v5shoutbox] Authenticated.");
      if(conn.caps.includes("sasl"))
        sendRaw("CAP END");
      setState(State.READY);
      conn.password = undefined;
      irc.onAuthenticated();

      if(conn.caps.includes("draft/chathistory")) {
        ["#annonces", "#projets", "#hs"].forEach(channel => {
          send("CHATHISTORY", ["LATEST", channel, "*", "100"]);
        });
      }
    }

    if(msg.command === "JOIN" && msg.tags.get("account") === conn.username) {
      irc.channels.set(msg.args[0], {});
      irc.onChannelChanged(msg.args[0]);
    }
    if(msg.command === 332) {
      irc.channels.get(msg.args[1]).header = msg.args[2];
      irc.onChannelChanged(msg.args[1]);
    }
    if(msg.command === 353 && msg.args[1] === "=") {
      irc.channels.get(msg.args[2]).clients = msg.args[3].split(" ");
      irc.onChannelChanged(msg.args[2]);
    }

    if(msg.command === "PRIVMSG")
      processPRIVMSG(msg);
  }

  function processBatch(batch) {
    const batchArgs = batch[0].args.slice(1);
    if(batchArgs.length === 2 && batchArgs[0] === "chathistory") {
      batch.forEach(msg => {
        if(msg.command === "PRIVMSG")
          processPRIVMSG(msg);
      });
    }
  }

  function processPRIVMSG(msg) {
    if(msg.args.length == 2) {
      let source = msg.source;
      if(source.includes("!"))
        source = source.substr(0, source.indexOf("!"));
      let date = new Date();
      if(msg.tags.has("time"))
        date = new Date(msg.tags.get("time"));
      irc.onNewMessage(msg.args[0], date, source, msg.args[1]);
    }
  }

  this.postMessage = function(channel, message) {
    if(message.startsWith("/")) {
      sendRaw(message.substring(1));
    }
    else {
      send("PRIVMSG", [channel, message]);
      if(!conn.caps.includes("echo-message"))
        irc.onNewMessage(channel, new Date(), conn.username, message);
    }
  }
};

/* Shoutbox entry point and DOM manipulation. */
let shoutbox = new function() {

  const serverUrl = "wss://irc.planet-casio.com:443";
  const postUrl = undefined;
  const availableChannels = ["#annonces", "#projets", "#hs"];

  /* Original tab title */
  this.title = undefined;
  /* Whether we currently have focus */
  this.focused = true;
  /* Number of messages received since last losing focus */
  this.newMessages = 0;

  this.init = function(root) {
    this.title = document.title;

    /* Channel views */
    this.eChannels = root.querySelector(".channels");
    this.eChannelButtons = root.querySelector(".channel-buttons");
    /* Debugging tools */
    this.eLog = root.querySelector(".log");
    this.eStatus = root.querySelector(".status");
    /* Elements of the login form */
    this.eLoginForm = root.querySelector(".login-form");
    this.eConnect = root.querySelector(".connect");
    this.eLogin = root.querySelector(".login");
    this.ePassword = root.querySelector(".password");
    /* Elements of the shoutbox form */
    this.eShoutboxForm = root.querySelector(".shoutbox-form");
    this.eMessage = root.querySelector(".message");

    this.eLoginForm.addEventListener("submit", e => {
      e.preventDefault();
      shoutbox.connect(serverUrl, postUrl);
    });
    this.eShoutboxForm.addEventListener("submit", e => {
      e.preventDefault();
      shoutbox.post();
    });
    this.eChannelButtons.querySelectorAll("button").forEach(b => {
      b.addEventListener("click", () => {
        shoutbox.selectChannel(b.dataset.channel);
      });
    });
    root.querySelector(".show-log").addEventListener("click", () => {
      this.selectChannel("\\log");
    });

    window.addEventListener("focus", () => {
      shoutbox.focused = true;
      shoutbox.newMessages = 0;
      shoutbox.refreshTitle();
    });
    window.addEventListener("blur", () => {
      shoutbox.focused = false;
    });

    this.selectChannel("\\login");
    this.refreshView();
  }

  /*** IRC callbacks ***/

  function log(message) {
    /* TODO: Use a better text element x) */
    this.eLog.innerHTML += message + "\n";
  }
  this.log = log.bind(this);
  irc.onLog = log.bind(this);

  irc.onAuthenticated = function() {
    this.ePassword.value = "";
  }.bind(this);

  this.refreshTitle = function() {
    if(this.newMessages === 0)
      document.title = this.title;
    else
      document.title = "(" + this.newMessages.toString() + ") " + this.title;
  }

  this.refreshView = function() {
    const running = irc.isRunning();
    if(irc.isOnRemoteChannel()) {
      const c = irc.currentChannel();
      this.eStatus.textContent = c + ": " + irc.channels.get(c).header;
    }
    else {
      this.eStatus.textContent = irc.stateString();
    }
    this.eShoutboxForm.style.display = running ? "flex" : "none";

    const name = irc.currentChannel();
    for(const e of this.eChannels.children)
      e.style.display = (e.dataset.channel === name) ? "flex" : "none";
    for(const b of this.eChannelButtons.children) {
      if(b.dataset.channel === name)
        b.classList.add("current");
      else
        b.classList.remove("current");
    }
  };
  irc.onStateChanged = this.refreshView.bind(this);

  irc.onNewMessage = function(channel, date, author, message) {
    shoutbox.addNewMessage(channel, date, author, message);

    if(channel === irc.currentChannel() && !shoutbox.focused) {
      shoutbox.newMessages++;
      shoutbox.refreshTitle();
    }
    if(channel !== irc.currentChannel()) {
      shoutbox.setChannelBackgroundActivity(channel, true);
    }
  }

  irc.onChannelChanged = function(channel) {
    for(const [channel, info] of irc.channels) {
      this.createChannel(channel);
      if(irc.currentChannel() === channel)
        this.refreshView();
    }
  }.bind(this);

  /*** DOM manipulation ***/

  this.getChannelView = function(channel) {
    return Array.from(this.eChannels.children)
           .find(e => e.dataset.channel === channel);
  }

  this.getChannelButton = function(channel) {
    return Array.from(this.eChannelButtons.children)
           .find(e => e.dataset.channel === channel);
  }

  this.createChannel = function(channel) {
    if(!availableChannels.includes(channel))
      return;
    if(this.getChannelView(channel) !== undefined)
      return;

    let view = document.createElement("div");
    view.classList.add("channel");
    view.dataset.channel = channel;
    view.style.display = "none";
    this.eChannels.appendChild(view);

    let button = document.createElement("button");
    button.appendChild(document.createTextNode(channel));
    button.dataset.channel = channel;
    this.eChannelButtons.appendChild(button);

    button.addEventListener("click", () => {
      shoutbox.selectChannel(button.dataset.channel);
    });

    if(irc.currentChannel() === "\\login")
      shoutbox.selectChannel(channel);
  }

  this.addNewMessage = function(channel, date, author, message) {
    const view = this.getChannelView(channel);
    if(view === undefined)
      return;

    /* Remove the shoutbox bridge's "[s]" suffix */
    if(author.endsWith("[s]"))
      author = author.substr(0, author.length - 3);

    let messageElement = document.createElement("div");
    messageElement.classList.add("message");
    let authorElement = document.createElement("div");
    authorElement.appendChild(document.createTextNode(author));
    authorElement.classList.add("message-author");
    messageElement.appendChild(authorElement);
    let messageContentElement = document.createElement("p");
    messageContentElement.classList.add("message-content");
    messageContentElement.appendChild(document.createTextNode(message));
    messageElement.appendChild(messageContentElement);
    let dateElement = document.createElement("div");
    dateElement.classList.add("message-date");
    dateElement.appendChild(document.createTextNode(date.toLocaleTimeString()));
    messageElement.appendChild(dateElement);
    view.prepend(messageElement);
  }.bind(this);

  this.setChannelBackgroundActivity = function(channel, activity) {
    const button = this.getChannelButton(channel);
    if(button !== undefined) {
      if(activity)
        button.classList.add("bg-activity");
      else
        button.classList.remove("bg-activity");
    }
  }

  /*** User interactions ***/

  this.connect = function(serverUrl, postUrl) {
    if(irc.isConnected())
      return this.log("Already connected!");
    if(this.eLogin.value === "" || this.ePassword.value === "")
      return this.log("Need login/password to connect!");
    if(this.eLogin.value.includes(" "))
      return this.log("Login should not contain a space!");

    irc.connect(serverUrl, postUrl, this.eLogin.value, this.ePassword.value);
  }

  this.selectChannel = function(name) {
    irc.selectChannel(name);
    this.setChannelBackgroundActivity(name, false);
    this.refreshView();
  }

  this.post = function() {
    if(this.eMessage.value === "")
      return;
    if(!irc.isRunning())
      return log("Cannot send message while not connected!");
    if(!irc.isOnRemoteChannel())
      return log("Cannot send message as no channel is selected!");

    irc.postMessage(irc.currentChannel(), this.eMessage.value);
    this.eMessage.value = "";
  }
};

document.addEventListener("DOMContentLoaded", function() {
  shoutbox.init(document.getElementById("v5shoutbox"));
});
